#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for openxpki_ca_handler"""

# pylint: disable=C0415, R0904, W0212
import sys
import os
import unittest
from unittest.mock import patch, Mock, MagicMock
import requests
import base64
from OpenSSL import crypto
import configparser

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class FakeDBStore(object):
    """face DBStore class needed for mocking"""

    # pylint: disable=W0107, R0903
    pass


class TestACMEHandler(unittest.TestCase):
    """test class for cgi_handler"""

    def setUp(self):
        """setup unittest"""
        models_mock = MagicMock()
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        import logging
        from acme2certifier.cahandlers.digicert_ca_handler import CAhandler

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        self.cahandler = CAhandler(False, self.logger)
        self.dir_path = os.path.dirname(os.path.realpath(__file__))

    def test_001_default(self):
        """default test which always passes"""
        self.assertEqual("foo", "foo")

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_load")
    def test_002__enter__(self, mock_cfg):
        """test enter  called"""
        mock_cfg.return_value = True
        self.cahandler.__enter__()
        self.assertTrue(mock_cfg.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_load")
    def test_003__enter__(self, mock_cfg):
        """test enter api hosts defined"""
        mock_cfg.return_value = True
        self.cahandler.api_key = "api_key"
        self.cahandler.__enter__()
        self.assertFalse(mock_cfg.called)

    def test_004_poll(self):
        """test polling"""
        self.assertEqual(
            ("Method not implemented.", None, None, "poll_identifier", False),
            self.cahandler.poll("cert_name", "poll_identifier", "csr"),
        )

    def test_005_trigger(self):
        """test polling"""
        self.assertEqual(
            ("Method not implemented.", None, None), self.cahandler.trigger("payload")
        )

    @patch.object(requests, "post")
    def test_006__api_post(self, mock_req):
        """test _api_post()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.json = lambda: {"foo": "bar"}
        mock_req.return_value = mockresponse
        self.assertEqual(
            ("status_code", {"foo": "bar"}), self.cahandler._api_post("url", "data")
        )

    @patch("requests.post")
    def test_007__api_post(self, mock_req):
        """test _api_post()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.json = "aaaa"
        mock_req.return_value = mockresponse
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("status_code", "'str' object is not callable"),
                self.cahandler._api_post("url", "data"),
            )
        self.assertIn(
            "ERROR:test_a2c:Request_operation returned error during json parsing: 'str' object is not callable",
            lcm.output,
        )

    @patch("requests.post")
    def test_008__api_post(self, mock_req):
        """test _api_post()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.text = None
        mock_req.return_value = mockresponse
        self.assertEqual(("status_code", None), self.cahandler._api_post("url", "data"))

    @patch("requests.post")
    def test_009__api_post(self, mock_req):
        """test _api_post(="""
        self.cahandler.api_host = "api_host"
        self.cahandler.auth = "auth"
        mock_req.side_effect = Exception("exc_api_post")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                (500, "exc_api_post"), self.cahandler._api_post("url", "data")
            )
        self.assertIn(
            "ERROR:test_a2c:Request_operation returned error: exc_api_post", lcm.output
        )

    @patch.object(requests, "get")
    def test_010__api_get(self, mock_req):
        """test _api_get()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.json = lambda: {"foo": "bar"}
        mock_req.return_value = mockresponse
        self.assertEqual(
            ("status_code", {"foo": "bar"}), self.cahandler._api_get("url")
        )

    @patch("requests.get")
    def test_011__api_get(self, mock_req):
        """test _api_get()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.json = "aaaa"
        mock_req.return_value = mockresponse
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("status_code", "'str' object is not callable"),
                self.cahandler._api_get("url"),
            )
        self.assertIn(
            "ERROR:test_a2c:Request_operation returned error during json parsing: 'str' object is not callable",
            lcm.output,
        )

    @patch("requests.get")
    def test_012__api_get(self, mock_req):
        """test _api_get()"""
        self.cahandler.api_host = "api_host"
        self.cahandler.auth = "auth"
        mock_req.side_effect = Exception("exc_api_get")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual((500, "exc_api_get"), self.cahandler._api_get("url"))
        self.assertIn(
            "ERROR:test_a2c:Request_operation returned error: exc_api_get", lcm.output
        )

    @patch.object(requests, "put")
    def test_013__api_put(self, mock_req):
        """test _api_put()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.json = lambda: {"foo": "bar"}
        mock_req.return_value = mockresponse
        self.assertEqual(
            ("status_code", {"foo": "bar"}), self.cahandler._api_put("url", "data")
        )

    @patch("requests.put")
    def test_014__api_put(self, mock_req):
        """test _api_put()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.json = "aaaa"
        mock_req.return_value = mockresponse
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("status_code", "'str' object is not callable"),
                self.cahandler._api_put("url", "data"),
            )
        self.assertIn(
            "ERROR:test_a2c:Request_operation returned error during json parsing: 'str' object is not callable",
            lcm.output,
        )

    @patch("requests.put")
    def test_015__api_put(self, mock_req):
        """test _api_put()"""
        mockresponse = Mock()
        mockresponse.status_code = "status_code"
        mockresponse.text = None
        mock_req.return_value = mockresponse
        self.assertEqual(("status_code", None), self.cahandler._api_put("url", "data"))

    @patch("requests.put")
    def test_016__api_put(self, mock_req):
        """test _api_put()"""
        self.cahandler.api_host = "api_host"
        self.cahandler.auth = "auth"
        mock_req.side_effect = Exception("exc_api_put")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                (500, "exc_api_put"), self.cahandler._api_put("url", "data")
            )
        self.assertIn(
            "ERROR:test_a2c:Request_operation returned error: exc_api_put", lcm.output
        )

    def test_017__config_check(self):
        """test _config_check()"""
        self.cahandler.api_url = "api_url"
        self.assertEqual(
            "api_key parameter is missing in config file",
            self.cahandler._config_check(),
        )

    def test_018__config_check(self):
        """test _config_check()"""
        self.cahandler.api_url = "api_url"
        self.cahandler.api_key = "api_key"
        self.assertEqual(
            "organization_name parameter is missing in config file",
            self.cahandler._config_check(),
        )

    def test_019__config_check(self):
        """test _config_check()"""
        self.cahandler.api_url = "api_url"
        self.cahandler.api_key = "api_key"
        self.cahandler.organization_name = "organization_name"
        self.assertFalse(self.cahandler._config_check())

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_020_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        # parser['CAhandler'] = {'foo': 'bar'}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_021_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_url": "api_url"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual("api_url", self.cahandler.api_url)
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_022_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_key": "api_key"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertEqual("api_key", self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_023_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"signature_hash": "signature_hash"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("signature_hash", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_024_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"cert_type": "cert_type"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("cert_type", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_025_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"order_validity": "2"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(2, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_026_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"order_validity": "aa"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertIn(
            "ERROR:test_a2c:Could not load order_validity:invalid literal for int() with base 10: 'aa'",
            lcm.output,
        )
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_027_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"request_timeout": 20}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(20, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_028_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"request_timeout": "30"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(30, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_029_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"request_timeout": "aa"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertIn(
            "ERROR:test_a2c:Could not load request_timeout:invalid literal for int() with base 10: 'aa'",
            lcm.output,
        )
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_030_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"organization_name": "organization_name"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertEqual("organization_name", self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_031_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"organization_id": "organization_id"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertEqual("organization_id", self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_032_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"allowed_domainlist": '["foo", "bar"]'}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_033_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"allowed_domainlist": '["foo"]'}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_headerinfo_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.load_config")
    def test_034_config_load(self, mock_load, mock_eab, mock_hdl):
        """test _config_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"allowed_domainlist": "foo"}
        mock_load.return_value = parser
        mock_eab.return_value = True, "eab"
        mock_hdl.return_value = "hdl"
        self.cahandler._config_load()
        self.assertTrue(mock_load.called)
        self.assertEqual(
            "https://www.digicert.com/services/v2/", self.cahandler.api_url
        )
        self.assertFalse(self.cahandler.api_key)
        self.assertEqual("ssl_basic", self.cahandler.cert_type)
        self.assertEqual("sha256", self.cahandler.signature_hash)
        self.assertEqual(1, self.cahandler.order_validity)
        self.assertEqual(10, self.cahandler.request_timeout)
        self.assertFalse(self.cahandler.organization_name)
        self.assertFalse(self.cahandler.organization_id)
        self.assertTrue(self.cahandler.eab_profiling)
        self.assertEqual("eab", self.cahandler.eab_handler)
        self.assertEqual("hdl", self.cahandler.header_info_field)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_035_order_send(self, mock_post, mock_orgid):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.organization_id = "organization_id"
        self.assertEqual(("code", "content"), self.cahandler._order_send("csr", "cn"))
        self.assertFalse(mock_orgid.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_036_order_send(self, mock_post, mock_orgid):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.assertEqual(
            (500, "organisation_id is missing"), self.cahandler._order_send("csr", "cn")
        )
        self.assertFalse(mock_orgid.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_037_order_send(self, mock_post, mock_orgid):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.eab_profiling = True
        self.cahandler.api_key = "api_key"
        self.cahandler.organization_name = "organization_name"
        self.cahandler.organization_id = "organization_id"
        self.assertEqual(("code", "content"), self.cahandler._order_send("csr", "cn"))
        self.assertTrue(mock_orgid.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_038_order_send(self, mock_post, mock_orgid):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.eab_profiling = True
        self.cahandler.api_key = "api_key"
        self.cahandler.organization_name = "organization_name"
        self.cahandler.organization_id = None
        mock_orgid.return_value = 1
        self.assertEqual(("code", "content"), self.cahandler._order_send("csr", "cn"))
        self.assertTrue(mock_orgid.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_039_order_send(self, mock_post, mock_orgid):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.api_key = "api_key"
        self.cahandler.organization_name = "organization_name"
        self.cahandler.organization_id = None
        mock_orgid.return_value = 1
        self.assertEqual(("code", "content"), self.cahandler._order_send("csr", "cn"))
        self.assertTrue(mock_orgid.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_040_order_send(self, mock_post, mock_orgid):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.api_key = "api_key"
        self.cahandler.organization_name = None
        self.cahandler.organization_id = None
        mock_orgid.return_value = 1
        self.assertEqual(
            (500, "organisation_id is missing"), self.cahandler._order_send("csr", "cn")
        )
        self.assertFalse(mock_orgid.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.enrollment_config_log")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_041_order_send(self, mock_post, mock_orgid, mock_ecl):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.api_key = None
        self.cahandler.organization_name = "organization_name"
        self.cahandler.organization_id = None
        mock_orgid.return_value = 1
        self.assertEqual(
            (500, "organisation_id is missing"), self.cahandler._order_send("csr", "cn")
        )
        self.assertFalse(mock_orgid.called)
        self.assertFalse(mock_ecl.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.enrollment_config_log")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._organiation_id_get")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_post")
    def test_042_order_send(self, mock_post, mock_orgid, mock_ecl):
        """test _order_send()"""
        mock_post.return_value = ("code", "content")
        self.cahandler.api_key = None
        self.cahandler.organization_name = "organization_name"
        self.cahandler.organization_id = None
        self.cahandler.enrollment_config_log = True
        mock_orgid.return_value = 1
        self.assertEqual(
            (500, "organisation_id is missing"), self.cahandler._order_send("csr", "cn")
        )
        self.assertFalse(mock_orgid.called)
        self.assertTrue(mock_ecl.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_pem2der")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.b64_encode")
    def test_043_order_response_parse(self, mock_b64, mock_pem2der):
        """test _order_parse()"""
        content_dic = {
            "id": "id",
            "certificate_chain": [{"pem": "pem1"}, {"pem": "pem2"}, {"pem": "pem3"}],
        }
        mock_b64.return_value = "b64"
        self.assertEqual(
            ("pem1\npem2\npem3\n", "b64", "id"),
            self.cahandler._order_response_parse(content_dic),
        )

    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_pem2der")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.b64_encode")
    def test_044_order_response_parse(self, mock_b64, mock_pem2der):
        """test _order_parse()"""
        content_dic = {
            "id": "id",
            "cert_chain": [{"pem": "pem1"}, {"pem": "pem2"}, {"pem": "pem3"}],
        }
        mock_b64.return_value = "b64"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                (None, None, None), self.cahandler._order_response_parse(content_dic)
            )
        self.assertIn(
            "ERROR:test_a2c:Order response parsing failed: no certificate_chain in response",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_pem2der")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.b64_encode")
    def test_045_order_response_parse(self, mock_b64, mock_pem2der):
        """test _order_parse()"""
        content_dic = {
            "id": "id",
            "certificate_chain": [{"pem": "pem1"}, {"_pem": "pem2"}, {"pem": "pem3"}],
        }
        mock_b64.return_value = "b64"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("pem1\npem3\n", "b64", "id"),
                self.cahandler._order_response_parse(content_dic),
            )
        self.assertIn(
            "ERROR:test_a2c:Order response parsing failed: no pem in certificate_chain",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_pem2der")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.b64_encode")
    def test_046_order_response_parse(self, mock_b64, mock_pem2der):
        """test _order_parse()"""
        content_dic = {
            "_id": "id",
            "certificate_chain": [{"pem": "pem1"}, {"pem": "pem2"}, {"pem": "pem3"}],
        }
        mock_b64.return_value = "b64"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("pem1\npem2\npem3\n", "b64", None),
                self.cahandler._order_response_parse(content_dic),
            )
        self.assertIn(
            "ERROR:test_a2c:Polling_identifier generation failed: no id in response",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_get")
    def test_047_organiation_id_get(self, mock_get):
        """test _organiation_id_get()"""
        mock_get.return_value = (500, {"id": "id"})
        self.cahandler.organization_name = "organization_name"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._organiation_id_get()
        self.assertIn("ERROR:test_a2c:Could not get organization id.", lcm.output)
        self.assertFalse(self.cahandler.organization_id)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_get")
    def test_048_organiation_id_get(self, mock_get):
        """test _organiation_id_get()"""
        mock_get.return_value = (
            200,
            {
                "organizations": [
                    {"name": "name1", "id": "id1"},
                    {"name": "name2", "id": "id2"},
                    {"name": "name3", "id": "id3"},
                ]
            },
        )
        self.cahandler.organization_name = "name1"
        self.assertEqual("id1", self.cahandler._organiation_id_get())

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_get")
    def test_049_organiation_id_get(self, mock_get):
        """test _organiation_id_get()"""
        mock_get.return_value = (
            200,
            {
                "organizations": [
                    {"name": "name1", "id": "id1"},
                    {"name": "name2", "id": "id2"},
                    {"name": "name3", "id": "id3"},
                ]
            },
        )
        self.cahandler.organization_name = "name2"
        self.assertEqual("id2", self.cahandler._organiation_id_get())

    @patch("acme2certifier.cahandlers.digicert_ca_handler.eab_profile_header_info_check")
    def test_050_csr_check(self, mock_ehichk):
        """test _csr_check()"""
        mock_ehichk.return_value = "mock_hichk"
        self.assertEqual("mock_hichk", self.cahandler._csr_check("csr"))

    @patch("acme2certifier.cahandlers.digicert_ca_handler.eab_profile_header_info_check")
    def test_051_csr_check(self, mock_ehichk):
        """test _csr_check()"""
        mock_ehichk.return_value = False
        self.assertFalse(self.cahandler._csr_check("csr"))

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_response_parse")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_send")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.csr_cn_lookup")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._csr_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    def test_052_enroll(
        self, mock_cfgchk, mock_csrchk, mock_cnget, mock_ordersend, mock_orderparse
    ):
        """test enroll()"""
        mock_cfgchk.return_value = "mock_cfgchk"
        mock_csrchk.return_value = "mock_csrchk"
        mock_cnget.return_value = "cn"
        mock_ordersend.return_value = ("code", "content")
        mock_orderparse.return_value = ("pem", "b64", "id")
        self.assertEqual(
            ("mock_cfgchk", None, None, None), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_cfgchk.called)
        self.assertFalse(mock_csrchk.called)
        self.assertFalse(mock_cnget.called)
        self.assertFalse(mock_ordersend.called)
        self.assertFalse(mock_orderparse.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_response_parse")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_send")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.csr_cn_lookup")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._csr_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    def test_053_enroll(
        self, mock_cfgchk, mock_csrchk, mock_cnget, mock_ordersend, mock_orderparse
    ):
        """test enroll()"""
        mock_cfgchk.return_value = False
        mock_csrchk.return_value = "mock_csrchk"
        mock_cnget.return_value = "cn"
        mock_ordersend.return_value = ("code", "content")
        mock_orderparse.return_value = ("pem", "b64", "id")
        self.assertEqual(
            ("mock_csrchk", None, None, None), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_cfgchk.called)
        self.assertTrue(mock_csrchk.called)
        self.assertFalse(mock_cnget.called)
        self.assertFalse(mock_ordersend.called)
        self.assertFalse(mock_orderparse.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_response_parse")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_send")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.csr_cn_lookup")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._csr_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    def test_054_enroll(
        self, mock_cfgchk, mock_csrchk, mock_cnget, mock_ordersend, mock_orderparse
    ):
        """test enroll()"""
        mock_cfgchk.return_value = False
        mock_csrchk.return_value = False
        mock_cnget.return_value = "cn"
        mock_ordersend.return_value = ("code", "content")
        mock_orderparse.return_value = ("pem", "b64", "id")
        self.assertEqual(
            ("Error during order creation: code - content", None, None, None),
            self.cahandler.enroll("csr"),
        )
        self.assertTrue(mock_cfgchk.called)
        self.assertTrue(mock_csrchk.called)
        self.assertTrue(mock_cnget.called)
        self.assertTrue(mock_ordersend.called)
        self.assertFalse(mock_orderparse.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_response_parse")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_send")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.csr_cn_lookup")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._csr_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    def test_055_enroll(
        self, mock_cfgchk, mock_csrchk, mock_cnget, mock_ordersend, mock_orderparse
    ):
        """test enroll()"""
        mock_cfgchk.return_value = False
        mock_csrchk.return_value = False
        mock_cnget.return_value = "cn"
        mock_ordersend.return_value = (
            "code",
            {"errors": [{"code": "code", "message": "content"}]},
        )
        mock_orderparse.return_value = ("pem", "b64", "id")
        self.assertEqual(
            (
                "Error during order creation: code - [{'code': 'code', 'message': 'content'}]",
                None,
                None,
                None,
            ),
            self.cahandler.enroll("csr"),
        )
        self.assertTrue(mock_cfgchk.called)
        self.assertTrue(mock_csrchk.called)
        self.assertTrue(mock_cnget.called)
        self.assertTrue(mock_ordersend.called)
        self.assertFalse(mock_orderparse.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_response_parse")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._order_send")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.csr_cn_lookup")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._csr_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    def test_056_enroll(
        self, mock_cfgchk, mock_csrchk, mock_cnget, mock_ordersend, mock_orderparse
    ):
        """test enroll()"""
        mock_cfgchk.return_value = False
        mock_csrchk.return_value = False
        mock_cnget.return_value = "cn"
        mock_ordersend.return_value = (200, "content")
        mock_orderparse.return_value = ("pem", "b64", "id")
        self.assertEqual((False, "pem", "b64", "id"), self.cahandler.enroll("csr"))
        self.assertTrue(mock_cfgchk.called)
        self.assertTrue(mock_csrchk.called)
        self.assertTrue(mock_cnget.called)
        self.assertTrue(mock_ordersend.called)
        self.assertTrue(mock_orderparse.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_put")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_serial_get")
    def test_057_revoke(self, mock_serial, mock_put, mock_cfgchk):
        """test revoke()"""
        mock_serial.return_value = "serial"
        mock_put.return_value = ("code", "content")
        self.assertEqual(("code", None, "content"), self.cahandler.revoke("cert"))
        self.assertFalse(mock_cfgchk.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_put")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_serial_get")
    def test_058_revoke(self, mock_serial, mock_put, mock_cfgchk):
        """test revoke()"""
        mock_serial.return_value = "serial"
        mock_put.return_value = ("code", "content")
        self.cahandler.eab_profiling = True
        self.assertEqual(("code", None, "content"), self.cahandler.revoke("cert"))
        self.assertTrue(mock_cfgchk.called)

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_put")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_serial_get")
    def test_059_revoke(self, mock_serial, mock_put):
        """test revoke()"""
        mock_serial.return_value = None
        mock_put.return_value = ("code", "content")
        self.assertEqual(
            (500, None, "Failed to parse certificate serial"),
            self.cahandler.revoke("cert"),
        )

    @patch("acme2certifier.cahandlers.digicert_ca_handler.CAhandler._api_put")
    @patch("acme2certifier.cahandlers.digicert_ca_handler.cert_serial_get")
    def test_060_revoke(self, mock_serial, mock_put):
        """test revoke()"""
        mock_serial.return_value = "serial"
        mock_put.return_value = (204, "content")
        self.assertEqual((200, None, "content"), self.cahandler.revoke("cert"))

    @patch("acme2certifier.cahandlers.digicert_ca_handler.handler_config_check")
    def test_061_handler_check(self, mock_handler_check):
        """test handler_check"""
        mock_handler_check.return_value = "mock_handler_check"
        self.assertEqual("mock_handler_check", self.cahandler.handler_check())


if __name__ == "__main__":

    unittest.main()



# -*- coding: utf-8 -*-
"""Tests for selectable DB handler resolution."""

from __future__ import annotations

import importlib
import logging
import os
from typing import Dict

import pytest

from acme2certifier.acme_srv import db_handler as db_handler_mod


@pytest.fixture(autouse=True)
def _clear_db_handler_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("ACME_SRV_DB_HANDLER", raising=False)


def test_normalize_short_names() -> None:
    assert (
        db_handler_mod._normalize_handler("wsgi")
        == "acme2certifier.dbhandlers.wsgi_handler"
    )
    assert (
        db_handler_mod._normalize_handler("DJANGO")
        == "acme2certifier.dbhandlers.django_handler"
    )
    assert (
        db_handler_mod._normalize_handler("my.custom.handler") == "my.custom.handler"
    )


def test_cfg_handler_wins_over_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("ACME_SRV_DB_HANDLER", "django")
    config: Dict[str, Dict[str, str]] = {"DBhandler": {"handler": "wsgi"}}
    assert (
        db_handler_mod.resolve_db_handler_module(config)
        == "acme2certifier.dbhandlers.wsgi_handler"
    )


def test_env_used_when_cfg_has_no_handler(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("ACME_SRV_DB_HANDLER", "django")
    config: Dict[str, Dict[str, str]] = {"DBhandler": {"dbfile": "/tmp/x.db"}}
    assert (
        db_handler_mod.resolve_db_handler_module(config)
        == "acme2certifier.dbhandlers.django_handler"
    )


def test_default_is_wsgi() -> None:
    config: Dict[str, Dict[str, str]] = {"DBhandler": {}}
    assert (
        db_handler_mod.resolve_db_handler_module(config)
        == "acme2certifier.dbhandlers.wsgi_handler"
    )


def test_handler_module_wins_over_handler() -> None:
    config: Dict[str, Dict[str, str]] = {
        "DBhandler": {
            "handler": "django",
            "handler_module": "acme2certifier.dbhandlers.wsgi_handler",
        }
    }
    assert (
        db_handler_mod.resolve_db_handler_module(config)
        == "acme2certifier.dbhandlers.wsgi_handler"
    )


def test_load_wsgi_handler_exports_dbstore() -> None:
    loaded = db_handler_mod.load_db_handler_module({"DBhandler": {"handler": "wsgi"}})
    assert hasattr(loaded, "DBstore")
    assert loaded.__name__ == "acme2certifier.dbhandlers.wsgi_handler"


def test_log_active_db_handler(caplog: pytest.LogCaptureFixture) -> None:
    logger = logging.getLogger("test.db_handler_startup")
    with caplog.at_level(logging.INFO, logger="test.db_handler_startup"):
        db_handler_mod.log_active_db_handler(logger)
    assert any("Using DB handler" in rec.message for rec in caplog.records)


def test_package_db_handler_reexports_dbstore() -> None:
    # Module already imported at collection time with default wsgi backend.
    assert hasattr(db_handler_mod, "DBstore")
    assert callable(db_handler_mod.DBstore)


def test_wsgi_backend_module_exports_dbstore() -> None:
    mod = importlib.import_module("acme2certifier.dbhandlers.wsgi_handler")
    assert mod.DBstore is not None
    assert mod.DBstore.__module__ == "acme2certifier.dbhandlers.wsgi_handler"



# -*- coding: utf-8 -*-
"""ca handler for "NetGuard Certificate Lifecycle Manager" via REST-API class"""

from __future__ import print_function
import os
import time
import json
from typing import List, Tuple, Dict
import requests

# pylint: disable=e0401, r0913
from acme2certifier.acme_srv.helper import (
    b64_encode,
    b64_url_recode,
    build_pem_file,
    cert_serial_get,
    config_eab_profile_load,
    config_enroll_config_log_load,
    config_headerinfo_load,
    config_profile_load,
    convert_string_to_byte,
    eab_profile_header_info_check,
    eab_profile_revocation_check,
    enrollment_config_log,
    error_dic_get,
    header_info_get,
    load_config,
    parse_url,
    proxy_check,
    uts_now,
    uts_to_date_utc,
)


class CAhandler(object):
    """CA  handler"""

    def __init__(self, _debug=None, logger=None):
        self.logger = logger
        self.api_host = None
        self.nclm_version = None
        self.api_version = "/v2"
        self.ca_bundle = True
        self.credential_dic = {"api_user": None, "api_password": None}
        self.container_info_dic = {"name": None, "id": None}
        self.template_info_dic = {"name": None, "id": None}
        self.headers = None
        self.ca_name = None
        self.error = None
        self.wait_interval = 5
        self.proxy = None
        self.request_timeout = 20
        self.header_info_field = False
        self.eab_handler = None
        self.eab_profiling = False
        self.enrollment_config_log = False
        self.enrollment_config_log_skip_list = []
        self.profiles = {}
        self.profile_mapping_field = "template_name"

    def __enter__(self):
        """Makes CAhandler a Context Manager"""
        if not self.api_host:
            self._config_load()
            self._config_check()
        if not self.headers and not self.error:
            self._login()
        if not self.container_info_dic["id"] and not self.error:
            self._container_id_lookup()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _api_post(self, url: str, data: Dict[str, str]) -> Dict[str, str]:
        """generic wrapper for an API post call"""
        self.logger.debug("CAhandler._api_post()")
        try:
            response = requests.post(
                url=url,
                json=data,
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            )
            try:
                api_response = response.json()
            except Exception:
                api_response = {"status": response.status_code}
        except Exception as err_:
            self.logger.error("API POST request failed: %s", err_)
            api_response = str(err_)

        self.logger.debug("CAhandler._api_post() ended with: %s", api_response)
        return api_response

    def _ca_id_get(self, ca_list: Dict[str, str]) -> int:
        """get ca_id"""
        self.logger.debug("CAhandler._ca_id_get()")
        ca_id = None
        if "items" in ca_list:
            for ca_ in ca_list["items"]:
                # compare name or description field against config value
                if "name" in ca_ and ca_["name"] == self.ca_name:
                    # pylint: disable=R1723
                    if "id" in ca_:
                        ca_id = ca_["id"]
                        break
                    else:
                        self.logger.error("CA response missing policyLinkId field.")

        self.logger.debug("CAhandler._ca_id_get() with %s", ca_id)
        return ca_id

    def _ca_policylink_id_lookup(self) -> int:
        """lookup CA ID based on CA_name"""
        self.logger.debug("CAhandler._ca_policylink_id_lookup()")

        # query CAs
        ca_list = requests.get(
            f'{self.api_host}{self.api_version}/containers/{self.container_info_dic["id"]}/issuers',
            headers=self.headers,
            verify=self.ca_bundle,
            proxies=self.proxy,
            timeout=self.request_timeout,
        ).json()
        if "items" in ca_list:
            ca_id = self._ca_id_get(ca_list)
        else:
            # log error
            ca_id = None
            self.logger.error("No CAs found in issuer response.")

        if not ca_id:
            # log error
            self.logger.error("No policy link ID found for CA name: %s", self.ca_name)
        self.logger.debug("CAhandler._ca_policylink_id_lookup() ended with: %s", ca_id)
        return ca_id

    def _cert_enroll(self, csr: str, policylink_id: int) -> Tuple[str, str, str]:
        """enroll operation"""
        self.logger.debug("CAhandler._cert_enroll()")

        error = None
        cert_bundle = None
        cert_raw = None
        cert_id = None

        # post csr
        job_id = self._csr_post(csr, policylink_id)

        if job_id:
            cert_id = self._cert_id_get(job_id)
            if cert_id:
                error, cert_bundle, cert_raw = self._cert_bundle_build(cert_id)
            else:
                self.logger.error("Certificate ID lookup failed for job: %s", job_id)
                error = "Certifcate_id lookup failed"
        else:
            self.logger.error("Job ID lookup failed during certificate enrollment.")
            error = "job_id lookup failed"

        self.logger.debug("CAhandler._cert_enroll() ended with error: %s", error)
        return (error, cert_bundle, cert_raw, cert_id)

    def _csr_post(self, csr: str, policylink_id: int) -> Dict[str, str]:
        """post csr"""
        self.logger.debug("CAhandler._csr_post()")

        job_id = None
        # build_pem_file
        csr = build_pem_file(self.logger, None, csr, 64, True)
        csr = b64_encode(self.logger, convert_string_to_byte(csr))
        data_dic = {"allowDuplicateCn": True, "request": {"pkcs10": csr}}

        # add template if correctly configured
        if "id" in self.template_info_dic and self.template_info_dic["id"]:
            data_dic["template"] = {"id": self.template_info_dic["id"]}

        response = self._api_post(
            f"{self.api_host}{self.api_version}/containers/{self.container_info_dic['id']}/issuers/{policylink_id}/csr",
            data_dic,
        )

        if "id" in response:
            job_id = response["id"]

        self.logger.debug("CAhandler._csr_post() ended with: %s", job_id)
        return job_id

    def _issuer_certid_get(self, cert_dic: Tuple[str, str]) -> Tuple[str, bool]:
        """get cert id of issuer"""
        self.logger.debug("CAhandler._issuer_certid_get()")

        cert_id = None
        issuer_loop = False

        if (
            isinstance(cert_dic, dict)
            and "urls" in cert_dic
            and "issuer" in cert_dic["urls"]
        ):
            self.logger.debug(
                "CAhandler._cert_bundle_build() fetch issuer : %s",
                cert_dic["urls"]["issuer"],
            )
            cert_dic = requests.get(
                self.api_host + cert_dic["urls"]["issuer"],
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
            if "urls" in cert_dic and "certificate" in cert_dic["urls"]:
                cert_id = cert_dic["urls"]["certificate"].replace(
                    "/v2/certificates/", ""
                )
                self.logger.debug(
                    "CAhandler._cert_bundle_build() fetch certificate for issuer-certid: %s",
                    cert_id,
                )
                issuer_loop = True

        self.logger.debug("CAhandler._issuer_certid_get() ended with: %s", cert_id)
        return (cert_id, issuer_loop)

    def _cert_bundle_build(self, cert_id: int) -> Tuple[str, str, str]:
        """download cert and create bundle"""
        self.logger.debug("CAhandler._cert_bundle_build(%s)", cert_id)
        cert_bundle = ""
        error = None
        cert_raw = None
        issuer_loop = True
        count = 0

        while issuer_loop:
            # set issuer loop to False to avoid ending in an endless loop
            issuer_loop = False
            count += 1
            self.logger.debug(
                "CAhandler._cert_bundle_build() fetch certificate for certid: %s",
                cert_id,
            )

            cert_dic = requests.get(
                f"{self.api_host}{self.api_version}/certificates/{cert_id}",
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
            if "der" in cert_dic:
                if count == 1:
                    # get cert_raw
                    cert_raw = cert_dic["der"]

                # build_pem_file
                cert_bundle = build_pem_file(
                    self.logger,
                    existing=cert_bundle,
                    certificate=cert_dic["der"],
                    wrap=True,
                    csr=False,
                )
                cert_id, issuer_loop = self._issuer_certid_get(cert_dic)

        # we need this for backwards compability
        if cert_bundle == "":
            cert_bundle = None

        self.logger.debug("CAhandler._cert_bundle_build() ended")
        return (error, cert_bundle, cert_raw)

    def _cert_id_get(self, job_id: int) -> int:
        """lookup get cert_id from enrollment job"""
        self.logger.debug("CAhandler._cert_id_get(%s)", job_id)

        cert_id = None
        # check job status
        cnt = 0
        while cnt < 10:
            response = requests.get(
                f"{self.api_host}{self.api_version}/jobs/{job_id}",
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
            if response.get("status", None) == "done":
                if (
                    len(response.get("entities", [])) > 0
                    and "ref" in response["entities"][0]
                    and response["entities"][0]["ref"].lower() == "certificate"
                    and "url" in response["entities"][0]
                ):
                    cert_id = response["entities"][0]["url"].replace(
                        "/v2/certificates/", ""
                    )
                    break
                else:
                    self.logger.error(
                        "Job completed but certificate reference is missing or malformed: %s",
                        response,
                    )
                    break

            self.logger.debug(
                "CAhandler._cert_id_get() waiting for job to complete. Attempt: %s status: %s",
                cnt,
                response.get("status", None),
            )
            cnt += 1
            time.sleep(self.wait_interval)

        self.logger.debug("CAhandler._cert_id_get() ended with: %s", cert_id)
        return cert_id

    def _certid_get_from_serial(self, cert_raw: str) -> List[str]:
        """get certificates"""
        self.logger.debug("CAhandler._certid_get_from_serial()")

        cert_serial = cert_serial_get(self.logger, cert_raw, hexformat=True)

        # search for certificate
        try:
            cert_list = requests.get(
                f"{self.api_host}{self.api_version}/certificates?freeText=={cert_serial}&containerId={self.container_info_dic['id']}",
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        except Exception as err:
            self.logger.error(
                "API request to fetch certificates got aborted with err: %s",
                err,
            )
            cert_list = []

        if (
            cert_list
            and "items" in cert_list
            and len(cert_list["items"]) > 0
            and "id" in cert_list["items"][0]
        ):
            cert_id = cert_list["items"][0]["id"]
        else:
            cert_id = None
            self.logger.error(
                "Failed to retrieve certificate by serial: %s",
                cert_serial,
            )

        self.logger.debug(
            "CAhandler._certid_get_from_serial() ended with code: %s", cert_id
        )
        return cert_id

    def _cert_id_lookup(self, cert_raw: str) -> int:
        """get tracking id"""
        self.logger.debug("CAhandler._cert_id_lookup()")

        cert_id = None

        # we misuse header_info_get() to get the tracking id from database
        cert_recode = b64_url_recode(self.logger, cert_raw)
        pid_list = header_info_get(
            self.logger,
            csr=cert_recode,
            vlist=["poll_identifier"],
            field_name="cert_raw",
        )

        for ele in pid_list:
            if "poll_identifier" in ele:
                cert_id = ele["poll_identifier"]
                break

        if not cert_id:
            # lookup through NCLM API
            self.logger.info("Cert_id not found in database. Lookup trough NCLM API")
            cert_id = self._certid_get_from_serial(cert_raw)

        self.logger.debug("CAhandler._cert_id_lookup() ended with %s", cert_id)
        return cert_id

    def _config_api_access_check(self):
        """check config for consitency"""
        self.logger.debug("CAhandler._config_api_access_check()")

        if not self.api_host:
            self.logger.error('Missing "api_host" in configuration.')
            self.error = "api_host to be set in config file"

        if not self.error and not self.credential_dic.get("api_user"):
            self.logger.error('Missing "api_user" in configuration.')
            self.error = "api_user to be set in config file"

        if not self.error and not (
            "api_password" in self.credential_dic
            and self.credential_dic["api_password"]
        ):
            self.logger.error('Missing "api_password" in configuration.')
            self.error = "api_password to be set in config file"

        self.logger.debug("CAhandler._config_api_access_check() ended")

    def _config_names_check(self):
        """check config for consitency"""
        self.logger.debug("CAhandler._config_names_check()")

        if not self.error and not self.container_info_dic.get("name"):
            self.logger.error('"tsg_name" to be set in config file')
            self.error = "tsg_name to be set in config file"

        if not self.error and not self.ca_name:
            self.logger.error('"ca_name" to be set in config file')
            self.error = "ca_name to be set in config file"

        if not self.error and self.ca_bundle is False:
            self.logger.warning(
                'CA bundle validation is disabled ("ca_bundle" set to False). Server certificate will not be validated.'
            )

        self.logger.debug("CAhandler._config_names_check() ended")

    def _config_check(self):
        """check config for consitency"""
        self.logger.debug("CAhandler._config_check()")

        self._config_api_access_check()
        self._config_names_check()

        self.logger.debug("CAhandler._config_check() ended")

    def _config_api_user_load(self, config_dic: Dict[str, str]):
        """load user"""
        self.logger.debug("CAhandler._config_api_user_load()")

        if "api_user_variable" in config_dic["CAhandler"]:
            try:
                self.credential_dic["api_user"] = os.environ[
                    config_dic.get("CAhandler", "api_user_variable")
                ]
            except Exception as err:
                self.logger.error("Unable to load API user from environment: %s", err)
        if "api_user" in config_dic["CAhandler"]:
            if self.credential_dic["api_user"]:
                self.logger.info("Overwrite api_user")
            self.credential_dic["api_user"] = config_dic.get("CAhandler", "api_user")

        self.logger.debug("CAhandler._config_api_user_load() ended.")

    def _config_api_password_load(self, config_dic: Dict[str, str]):
        """load password"""
        self.logger.debug("CAhandler._config_api_password_load()")

        if "api_password_variable" in config_dic["CAhandler"]:
            try:
                self.credential_dic["api_password"] = os.environ[
                    config_dic.get("CAhandler", "api_password_variable")
                ]
            except Exception as err:
                self.logger.error("Could not load password_variable:%s", err)
        if "api_password" in config_dic["CAhandler"]:
            if self.credential_dic["api_password"]:
                self.logger.info("Overwrite api_password")
            self.credential_dic["api_password"] = config_dic.get(
                "CAhandler", "api_password"
            )

        self.logger.debug("CAhandler._config_api_password_load() ended")

    def _config_names_load(self, config_dic: Dict[str, str]):
        """load names from config"""
        self.logger.debug("CAhandler._config_names_load()")

        self.api_host = config_dic.get("CAhandler", "api_host", fallback=self.api_host)
        self.ca_name = config_dic.get("CAhandler", "ca_name", fallback=self.ca_name)
        self.template_info_dic["name"] = config_dic.get(
            "CAhandler", self.profile_mapping_field, fallback=None
        )
        if "container_name" in config_dic["CAhandler"]:
            self.container_info_dic["name"] = config_dic.get(
                "CAhandler", "container_name", fallback=None
            )
        elif "tsg_name" in config_dic["CAhandler"]:
            # for backwards compatibility
            self.logger.warning(
                "Configuration uses deprecated 'tsg_name'. Use 'container_name' instead."
            )
            self.container_info_dic["name"] = config_dic.get(
                "CAhandler", "tsg_name", fallback=None
            )

        self.logger.debug("CAhandler._config_names_load() ended")

    def _config_proxy_load(self, config_dic: Dict[str, str]):
        """load proxy configuration"""
        self.logger.debug("CAhandler._config_proxy_load()")

        if "DEFAULT" in config_dic and "proxy_server_list" in config_dic["DEFAULT"]:
            try:
                proxy_list = json.loads(config_dic.get("DEFAULT", "proxy_server_list"))
                url_dic = parse_url(self.logger, self.api_host)
                if "host" in url_dic:
                    fqdn, _port = url_dic["host"].split(":")
                    proxy_server = proxy_check(self.logger, fqdn, proxy_list)
                    self.proxy = {"http": proxy_server, "https": proxy_server}
            except Exception as err_:
                self.logger.warning(
                    "Failed to load proxy_server_list from configuration: %s",
                    err_,
                )
        self.logger.debug("CAhandler._config_proxy_load() ended")

    def _config_timer_load(self, config_dic: Dict[str, str]):
        """load timer"""
        self.logger.debug("CAhandler._config_proxy_load()")

        # check if we get a ca bundle for verification
        if "ca_bundle" in config_dic["CAhandler"]:
            try:
                self.ca_bundle = config_dic.getboolean("CAhandler", "ca_bundle")
            except Exception:
                self.ca_bundle = config_dic.get(
                    "CAhandler", "ca_bundle", fallback=self.ca_bundle
                )

        if "request_timeout" in config_dic["CAhandler"]:
            try:
                self.request_timeout = int(
                    config_dic.get(
                        "CAhandler", "request_timeout", fallback=self.request_timeout
                    )
                )
            except Exception:
                self.request_timeout = 20

        self.logger.debug("CAhandler._config_proxy_load() ended")

    def _config_load(self):
        """ " load config from file"""
        # pylint: disable=r0912
        self.logger.debug("CAhandler._config_load()")
        config_dic = load_config(self.logger, "CAhandler")
        if "CAhandler" in config_dic:

            self._config_names_load(config_dic)
            self._config_api_user_load(config_dic)
            self._config_api_password_load(config_dic)
            self._config_timer_load(config_dic)

        self._config_proxy_load(config_dic)
        # load profiling
        self.eab_profiling, self.eab_handler = config_eab_profile_load(
            self.logger, config_dic
        )
        # load profiles
        self.profiles = config_profile_load(self.logger, config_dic)
        # load header info
        self.header_info_field = config_headerinfo_load(self.logger, config_dic)
        # load enrollment config log
        (
            self.enrollment_config_log,
            self.enrollment_config_log_skip_list,
        ) = config_enroll_config_log_load(self.logger, config_dic)

        self.logger.debug("CAhandler._config_load() ended")

    def _container_id_lookup(self):
        """get target system id based on name"""
        self.logger.debug(
            "CAhandler._container_id_lookup() for tsg: %s",
            self.container_info_dic["name"],
        )
        try:
            tsg_list = requests.get(
                self.api_host
                + "/containers?freeText="
                + str(self.container_info_dic["name"])
                + "&offset=0&limit=50&fetchPath=true",
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        except Exception as err_:
            self.logger.error("Failed to retrieve container id: %s", err_)
            tsg_list = []

        if "items" in tsg_list:
            for tsg in tsg_list["items"]:
                if "name" in tsg and "id" in tsg:
                    if self.container_info_dic["name"] == tsg["name"]:
                        self.container_info_dic["id"] = tsg["id"]
                        break
                else:
                    self.logger.error("Incomplete container response: %s", tsg)
        else:
            self.logger.error(
                "No target system groups found for filter: %s.",
                self.container_info_dic["name"],
            )
        self.logger.debug(
            "CAhandler._container_id_lookup() ended with: %s",
            str(self.container_info_dic["id"]),
        )

    def _csr_check(self, csr: str) -> str:
        """check csr"""
        self.logger.debug("CAhandler._csr_check()")

        # check for eab profiling and header_info
        error = eab_profile_header_info_check(
            self.logger, self, csr, self.profile_mapping_field
        )

        self.logger.debug("CAhandler._csr_check() ended with: %s", error)
        return error

    def _enroll(self, csr: str, ca_id: int) -> Tuple[str, str, str, str]:
        """enroll certificate from NCLM"""
        self.logger.debug("CAhandler._enroll()")

        error = None
        cert_bundle = None
        cert_raw = None
        cert_id = None

        if self.enrollment_config_log:
            self.enrollment_config_log_skip_list.extend(["headers", "credential_dic"])
            enrollment_config_log(
                self.logger, self, self.enrollment_config_log_skip_list
            )

        if ca_id and self.container_info_dic["id"]:
            # enroll operation
            error, cert_bundle, cert_raw, cert_id = self._cert_enroll(csr, ca_id)
        else:
            error = f'Enrollment aborted. ca: {ca_id}, tsg_id: {self.container_info_dic["id"]}'
            self.logger.info(
                "CAhandler.eroll(): Enrollment aborted. ca_id: %s, container: %s",
                ca_id,
                self.container_info_dic["id"],
            )

        self.logger.debug("CAhandler._enroll() ended with: %s", error)
        return (error, cert_bundle, cert_raw, cert_id)

    def _login(self):
        """_login into NCLM API"""
        self.logger.debug("CAhandler._login()")
        # check first if API is reachable
        api_response = requests.get(
            self.api_host + "/v1",
            proxies=self.proxy,
            timeout=self.request_timeout,
            verify=self.ca_bundle,
        )
        self.logger.debug("api response code:%s", api_response.status_code)

        if api_response.ok:
            # all fine try to login
            if "versionNumber" in api_response.json():
                self.nclm_version = api_response.json()["versionNumber"]
                self.logger.debug("NCLM version: %s", self.nclm_version)

            self.logger.debug(
                'log in to %s as user "%s"',
                self.api_host,
                self.credential_dic["api_user"],
            )
            data = {
                "username": self.credential_dic["api_user"],
                "password": self.credential_dic["api_password"],
            }
            api_response = requests.post(
                url=self.api_host
                + self.api_version
                + "/token?grant_type=client_credentials",
                json=data,
                proxies=self.proxy,
                timeout=self.request_timeout,
                verify=self.ca_bundle,
            )

            if api_response.ok:
                json_dic = api_response.json()
                if "access_token" in json_dic:
                    self.headers = {
                        "Authorization": f"Bearer {json_dic['access_token']}"
                    }
                    _username = json_dic.get("username", None)
                    _realms = json_dic.get("realms", None)
                    self.logger.debug(
                        "login response:\n user: %s\n token: %s\n realms: %s\n",
                        _username,
                        json_dic["access_token"],
                        _realms,
                    )
                else:
                    self.logger.error("No token returned after logging in. Aborting.")
            else:
                self.logger.error("Login Error: %s", api_response.status_code)
        else:
            # If response code is not ok (200), print the resulting http error code with description
            self.logger.error("Login failed. Error: %s", api_response.status_code)

    def _revocation_status_poll(
        self, job_id: int, err_dic: Dict[str, str]
    ) -> Tuple[int, str, str]:
        """poll status of revocation job"""
        self.logger.debug("CAhandler._revocation_status_poll()")

        cnt = 0
        while cnt < 10:
            response = requests.get(
                f"{self.api_host}{self.api_version}/jobs/{job_id}",
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
            if "status" in response and response["status"] in ["done", "failed"]:
                if response["status"] == "done":
                    code = 200
                    message = None
                    detail = None
                elif response["status"] == "failed":
                    code = 500
                    message = err_dic["serverinternal"]
                    detail = "Revocation operation failed: error from API"
                break
            time.sleep(self.wait_interval)
            cnt += 1

        if cnt == 10:
            code = 500
            message = err_dic["serverinternal"]
            detail = "Revocation operation failed: Timeout"

        self.logger.debug("CAhandler._revocation_status_poll() ended with: %s", code)
        return (code, message, detail)

    def _template_list_get(self, ca_id: int) -> Dict[str, str]:
        """get list of templates"""
        self.logger.debug("CAhandler._template_list_get(%s)", ca_id)
        try:
            template_list = requests.get(
                f"{self.api_host}{self.api_version}/containers/{self.container_info_dic['id']}/issuers/{ca_id}/templates",
                headers=self.headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        except Exception as err_:
            self.logger.error("Failed to retrieve template list: %s", err_)
            template_list = []

        if "items" in template_list:
            tmpl_cnt = len(template_list["items"])
        else:
            tmpl_cnt = 0

        self.logger.debug(
            "CAhandler._template_list_get() ended with: %s templates", tmpl_cnt
        )
        return template_list

    def _templates_enumerate(self, template_list: Dict[str, str]):
        """get template id based on name"""
        self.logger.debug(
            "CAhandler._templates_enumerate() for template: %s",
            self.template_info_dic["name"],
        )

        for template in template_list["items"]:
            if (
                "name" in template
                and template["name"] == self.template_info_dic["name"]
                and "id" in template
            ):
                self.template_info_dic["id"] = template["id"]
                break

    def _template_id_lookup(self, ca_id: int):
        """get template id based on name"""
        self.logger.debug(
            "CAhandler._template_id_lookup() for template: %s",
            self.template_info_dic["name"],
        )

        # get list of templates
        template_list = self._template_list_get(ca_id)

        # enumerate templates to get template-id
        if "items" in template_list:
            self._templates_enumerate(template_list)
        else:
            self.logger.error(
                "No templates found for filter: %s.",
                self.template_info_dic["name"],
            )

        self.logger.debug(
            "CAhandler._template_id_lookup() ended with: %s",
            str(self.template_info_dic["id"]),
        )

    def enroll(self, csr: str) -> Tuple[str, str, str, str]:
        """enroll certificate from NCLM"""
        self.logger.debug("CAhandler.enroll()")

        cert_bundle = None
        error = None
        cert_raw = None
        cert_id = None

        # recode csr
        csr = b64_url_recode(self.logger, csr)

        if not self.error:
            if self.container_info_dic["id"]:

                # templating
                ca_id = self._ca_policylink_id_lookup()

                if (
                    ca_id
                    and self.template_info_dic["name"]
                    and not self.template_info_dic["id"]
                ):
                    self._template_id_lookup(ca_id)

                error = self._csr_check(csr)

                if not error:
                    error, cert_bundle, cert_raw, cert_id = self._enroll(csr, ca_id)
                else:
                    self.logger.error(
                        "EAB profile lookup failed with error: %s",
                        error,
                    )
            else:
                error = f'ID lookup for container"{self.container_info_dic["name"]}" failed.'
        else:
            error = self.error
            self.logger.error(self.error)

        self.logger.debug("CAhandler.enroll() ended")
        return (error, cert_bundle, cert_raw, cert_id)

    def handler_check(self):
        """check if handler is ready"""
        self.logger.debug("CAhandler.check()")
        self._config_check()

        self.logger.debug("CAhandler.check() ended with %s", self.error)
        return self.error

    def poll(
        self, _cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[str, str, str, str, bool]:
        """poll status of pending CSR and download certificates"""
        self.logger.debug("CAhandler.poll()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None
        rejected = False

        self.logger.debug("CAhandler.poll() ended")
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def revoke(
        self,
        cert: str,
        rev_reason: str = "unspecified",
        rev_date: str = uts_to_date_utc(uts_now()),
    ) -> Tuple[int, str, str]:
        """revoke certificate"""
        self.logger.debug("CAhandler.revoke()")

        err_dic = error_dic_get(self.logger)

        code = 500
        message = err_dic["serverinternal"]
        detail = "Revocation operation failed"

        # modify handler configuration in case of eab profiling
        if self.eab_profiling:
            eab_profile_revocation_check(self.logger, self, cert)

        # get tracking id as input for revocation call
        cert_id = self._cert_id_lookup(cert)

        if cert_id:
            data_dic = {"reason": rev_reason, "time": rev_date}
            response = self._api_post(
                f"{self.api_host}{self.api_version}/certificates/{cert_id}/revoke",
                data_dic,
            )
            if "urls" in response and "job" in response["urls"]:
                job_id = response["urls"]["job"].replace("/v2/jobs/", "")
            else:
                job_id = None
                self.logger.error(
                    "Job ID lookup failed for certificate: %s",
                    cert_id,
                )

        if job_id:
            code, message, detail = self._revocation_status_poll(job_id, err_dic)

        self.logger.debug("CAhandler.revoke() ended with: %s", code)
        return (code, message, detail)

    def trigger(self, _payload: str) -> Tuple[str, str, str]:
        """process trigger message and return certificate"""
        self.logger.debug("CAhandler.trigger()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)



"""
A Python client for the Microsoft AD Certificate Services web page.

https://github.com/magnuswatn/certsrv
"""

# pylint: disable=C0209, C0415, R1720, R1705
import os
import re
import base64
import logging
import warnings
import requests

__version__ = "2.1.1"

logger = logging.getLogger(__name__)

TIMEOUT = 30
UNKOWN_ERR_MSG = "An unknown error occured"
DEPRECATIONWARNING = (
    "This function is deprecated. Use the method on the Certsrv class instead"
)


class RequestDeniedException(Exception):
    """Signifies that the request was denied by the ADCS server."""

    def __init__(self, message, response):
        Exception.__init__(self, message)
        self.response = response


class CouldNotRetrieveCertificateException(Exception):
    """Signifies that the certificate could not be retrieved."""

    def __init__(self, message, response):
        Exception.__init__(self, message)
        self.response = response


class CertificatePendingException(Exception):
    """Signifies that the request needs to be approved by a CA admin."""

    def __init__(self, req_id):
        Exception.__init__(
            self,
            "Your certificate request has been received. "
            "However, you must wait for an administrator to issue the "
            "certificate you requested. Your Request Id is {0}.".format(req_id),
        )
        self.req_id = req_id


class Certsrv(object):
    """
    Represents a Microsoft AD Certificate Services web server.

    Args:
        server: The FQDN to a server running the Certification Authority
            Web Enrollment role (must be listening on https).
        username: The username for authentication.
        password: The password for authentication.
        auth_method: The chosen authentication method. Either 'basic' (the default),
            'ntlm', 'cert' (SSL client certificate) or 'gssapi' (GSSAPI, Kerberos)
        cafile: A PEM file containing the CA certificates that should be trusted.
        verify: Boolean to enable/disable CA certificate checking.
        timeout: The timeout to use against the CA server, in seconds.
            The default is 30.
        proxies: Dictionary of proxy server for post of get operations
            {'http': 'http://foo.bar:3128', 'https': 'socks5://foo.bar:1080'}
            The default is None
    Note:
        If you use a client certificate for authentication (auth_method=cert),
        the username parameter should be the path to a certificate, and
        the password parameter the path to a (unencrypted) private key.
    """

    # pylint: disable=r0913
    def __init__(
        self,
        server,
        url,
        username,
        password,
        auth_method="basic",
        cafile=None,
        verify=True,
        timeout=TIMEOUT,
        proxies=None,
    ):

        self.server = server
        self.url = url
        self.timeout = timeout
        self.auth_method = auth_method
        self.session = requests.Session()
        self.proxies = proxies

        if not verify:
            self.session.verify = False
        elif cafile:
            self.session.verify = cafile
        else:
            # requests uses it's own CA bundle by default
            # but ADCS servers often have certificates
            # from private CAs that are locally trusted,
            # so we try to find, and use, the system bundle
            # instead. Fallback to requests own.
            self.session.verify = _get_ca_bundle()

        self._set_credentials(username, password)

        # We need certsrv to think we are a browser,
        # or otherwise the Content-Type of the retrieved
        # certificate will be wrong (for some reason).
        self.session.headers = {
            "User-agent": "Mozilla/5.0 certsrv (https://github.com/magnuswatn/certsrv)"
        }

    def _set_credentials(self, username, password):
        if self.auth_method == "ntlm":
            from requests_ntlm import HttpNtlmAuth

            self.session.auth = HttpNtlmAuth(username, password)
        elif self.auth_method == "cert":
            self.session.cert = (username, password)
        elif self.auth_method == "gssapi":
            from requests_gssapi import HTTPSPNEGOAuth
            import gssapi

            # Support two GSSAPI modes:
            # 1) username/password (legacy behavior)
            # 2) default credential cache (for keytab/kinit prepared by caller)
            if password:
                oid = "1.3.6.1.5.5.2"  # SPNEGO
                # pylint: disable=e1101
                cred = gssapi.raw.acquire_cred_with_password(
                    gssapi.Name(username, gssapi.NameType.user),
                    password.encode("utf-8"),
                    mechs=[gssapi.OID.from_int_seq(oid)],
                    usage="initiate",
                )
                self.session.auth = HTTPSPNEGOAuth(
                    creds=cred.creds, mech=gssapi.OID.from_int_seq(oid)
                )
            else:
                self.session.auth = HTTPSPNEGOAuth()
        else:
            self.session.auth = (username, password)

    def _post(self, url, **kwargs):
        response = self.session.post(
            url, timeout=self.timeout, proxies=self.proxies, **kwargs
        )
        return self._handle_response(response)

    def _get(self, url, **kwargs):
        response = self.session.get(
            url, timeout=self.timeout, proxies=self.proxies, **kwargs
        )
        return self._handle_response(response)

    @staticmethod
    def _handle_response(response):

        logger.debug(
            "Sent %s request to %s, with headers:\n%s\n\nand body:\n%s",
            response.request.method,
            response.request.url,
            "\n".join(
                ["{0}: {1}".format(k, v) for k, v in response.request.headers.items()]
            ),
            response.request.body,
        )

        try:
            debug_content = response.content.decode()
        except UnicodeDecodeError:
            debug_content = base64.b64encode(response.content)

        logger.debug(
            "Recieved response:\nHTTP %s\n%s\n\n%s",
            response.status_code,
            "\n".join(["{0}: {1}".format(k, v) for k, v in response.headers.items()]),
            debug_content,
        )

        response.raise_for_status()

        return response

    def get_cert(self, csr, template, encoding="b64", attributes=None):
        """
        Gets a certificate from the ADCS server.

        Args:
            csr: The certificate request to submit.
            template: The certificate template the cert should be issued from.
            encoding: The desired encoding for the returned certificate.
                Possible values are 'bin' for binary and 'b64' for Base64 (PEM).
            attributes: Additional Attributes (request attibutes) to be sent along with
                the request.

        Returns:
            The issued certificate.

        Raises:
            RequestDeniedException: If the request was denied by the ADCS server.
            CertificatePendingException: If the request needs to be approved
                by a CA admin.
            CouldNotRetrieveCertificateException: If something went wrong while
                fetching the cert.
        """
        cert_attrib = "CertificateTemplate:{0}\r\n".format(template)
        if attributes:
            cert_attrib += attributes

        data = {
            "Mode": "newreq",
            "CertRequest": csr,
            "CertAttrib": cert_attrib,
            "FriendlyType": "Saved-Request Certificate",
            "TargetStoreFlags": "0",
            "SaveCert": "yes",
        }

        url = ""
        if self.url:
            url = "{0}/certfnsh.asp".format(self.url)
        else:
            url = "https://{0}/certsrv/certfnsh.asp".format(self.server)

        response = self._post(url, data=data)

        # We need to parse the Request ID from the returning HTML page
        try:
            req_id = re.search(r"certnew.cer\?ReqID=(\d+)&", response.text).group(1)
        except AttributeError:
            # We didn't find any request ID in the response. It may need approval.
            if re.search(r"Certificate Pending", response.text):
                req_id = re.search(r"Your Request Id is (\d+).", response.text).group(1)
                # pylint: disable=w0707
                raise CertificatePendingException(req_id)
            else:
                # Must have failed. Lets find the error message
                # and raise a RequestDeniedException.
                try:
                    error = re.search(
                        r'The disposition message is "([^"]+)', response.text
                    ).group(1)
                except AttributeError:
                    error = UNKOWN_ERR_MSG
                # pylint: disable=w0707
                raise RequestDeniedException(error, response.text)

        return self.get_existing_cert(req_id, encoding)

    def get_existing_cert(self, req_id, encoding="b64"):
        """
        Gets a certificate that has already been created from the ADCS server.

        Args:
            req_id: The request ID to retrieve.
            encoding: The desired encoding for the returned certificate.
                Possible values are 'bin' for binary and 'b64' for Base64 (PEM).

        Returns:
            The issued certificate.

        Raises:
            CouldNotRetrieveCertificateException: If something went wrong
                while fetching the cert.
        """

        cert_url = ""
        if self.url:
            cert_url = "{0}/certnew.cer".format(self.url)
        else:
            cert_url = "https://{0}/certsrv/certnew.cer".format(self.server)

        params = {"ReqID": req_id, "Enc": encoding}

        response = self._get(cert_url, params=params)

        if response.headers["Content-Type"] != "application/pkix-cert":
            # The response was not a cert. Something must have gone wrong
            try:
                error = re.search(
                    "Disposition message:[^\t]+\t\t([^\r\n]+)", response.text
                ).group(1)

            except AttributeError:
                error = UNKOWN_ERR_MSG
            raise CouldNotRetrieveCertificateException(error, response.text)
        else:
            return response.content

    def get_ca_cert(self, encoding="b64"):
        """
        Gets the (newest) CA certificate from the ADCS server.

        Args:
            encoding: The desired encoding for the returned certificate.
                Possible values are 'bin' for binary and 'b64' for Base64 (PEM).

        Returns:
            The newest CA certificate from the server.
        """
        url = ""
        if self.url:
            url = "{0}/certcarc.asp".format(self.url)
        else:
            url = "https://{0}/certsrv/certcarc.asp".format(self.server)

        response = self._get(url)

        # We have to check how many renewals this server has had,
        # so that we get the newest CA cert.
        renewals = re.search(r"var nRenewals=(\d+);", response.text).group(1)

        cert_url = ""
        if self.url:
            cert_url = "{0}/certnew.cer".format(self.url)
        else:
            cert_url = "https://{0}/certsrv/certnew.cer".format(self.server)
        params = {"ReqID": "CACert", "Enc": encoding, "Renewal": renewals}

        response = self._get(cert_url, params=params)

        if response.headers["Content-Type"] != "application/pkix-cert":
            raise CouldNotRetrieveCertificateException(UNKOWN_ERR_MSG, response.content)

        return response.content

    def get_chain(self, encoding="bin"):
        """
        Gets the CA chain from the ADCS server.

        Args:
            encoding: The desired encoding for the returned certificates.
                Possible values are 'bin' for binary and 'b64' for Base64 (PEM).

        Returns:
            The CA chain from the server, in PKCS#7 format.
        """
        url = ""
        if self.url:
            url = "{0}/certcarc.asp".format(self.url)
        else:
            url = "https://{0}/certsrv/certcarc.asp".format(self.server)

        response = self._get(url)

        # We have to check how many renewals this server has had, so that we get the newest chain
        renewals = re.search(r"var nRenewals=(\d+);", response.text).group(1)

        chain_url = ""
        if self.url:
            chain_url = "{0}/certnew.p7b".format(self.url)
        else:
            chain_url = "https://{0}/certsrv/certnew.p7b".format(self.server)

        params = {"ReqID": "CACert", "Renewal": renewals, "Enc": encoding}

        chain_response = self._get(chain_url, params=params)

        if chain_response.headers["Content-Type"] != "application/x-pkcs7-certificates":
            raise CouldNotRetrieveCertificateException(
                UNKOWN_ERR_MSG, chain_response.content
            )

        return chain_response.content

    def check_credentials(self):
        """
        Checks the specified credentials against the ADCS server.

        Returns:
            True if authentication succeeded, False if it failed.
        """
        url = ""
        if self.url:
            url = "{0}".format(self.url)
        else:
            url = "https://{0}/certsrv/".format(self.server)

        try:
            self._get(url)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 401:
                return False
            else:
                raise
        return True

    def update_credentials(self, username, password):
        """
        Updates the credentials used against the ADCS server.

        Args:
            username: The username for authentication.
            password: The password for authentication.
        """
        if self.auth_method in ("ntlm", "cert", "gssapi"):
            # NTLM and SSL is connection based,
            # so we need to close the connection
            # to be able to re-authenticate
            self.session.close()
        self._set_credentials(username, password)


def _get_ca_bundle():
    """Tries to find the platform ca bundle for the system (on linux systems)"""
    ca_bundles = [
        # list taken from https://golang.org/src/crypto/x509/root_linux.go
        "/etc/ssl/certs/ca-certificates.crt",  # Debian/Ubuntu/Gentoo etc.
        "/etc/pki/tls/certs/ca-bundle.crt",  # Fedora/RHEL 6
        "/etc/ssl/ca-bundle.pem",  # OpenSUSE
        "/etc/pki/tls/cacert.pem",  # OpenELEC
        "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",  # CentOS/RHEL 7
    ]
    for ca_bundle in ca_bundles:
        if os.path.isfile(ca_bundle):
            return ca_bundle
    # if the bundle was not found, we revert back to requests own
    return True


# pylint: disable=r0913
def get_cert(server, csr, template, username, password, encoding="b64", **kwargs):
    """
    Gets a certificate from a Microsoft AD Certificate Services web page.

    Args:
        server: The FQDN to a server running the Certification Authority
            Web Enrollment role (must be listening on https).
        csr: The certificate request to submit.
        template: The certificate template the cert should be issued from.
        username: The username for authentication.
        pasword: The password for authentication.
        encoding: The desired encoding for the returned certificate.
            Possible values are 'bin' for binary and 'b64' for Base64 (PEM).
        auth_method: The chosen authentication method. Either 'basic' (the default),
            'ntlm', 'cert' (ssl client certificate) or 'gssapi' (GSSAPI, Kerberos).
        cafile: A PEM file containing the CA certificates that should be trusted.

    Returns:
        The issued certificate.

    Raises:
        RequestDeniedException: If the request was denied by the ADCS server.
        CertificatePendingException: If the request needs to be approved by a CA admin.
        CouldNotRetrieveCertificateException: If something went wrong while
            fetching the cert.

    Note:
        This method is deprecated.

    """
    warnings.warn(
        DEPRECATIONWARNING,
        DeprecationWarning,
    )
    certsrv = Certsrv(server, username, password, **kwargs)
    return certsrv.get_cert(csr, template, encoding)


def get_existing_cert(server, req_id, username, password, encoding="b64", **kwargs):
    """
    Gets a certificate that has already been created from a
    Microsoft AD Certificate Services web page.

    Args:
        server: The FQDN to a server running the Certification Authority
            Web Enrollment role (must be listening on https).
        req_id: The request ID to retrieve.
        username: The username for authentication.
        pasword: The password for authentication.
        encoding: The desired encoding for the returned certificate.
            Possible values are 'bin' for binary and 'b64' for Base64 (PEM).
        auth_method: The chosen authentication method. Either 'basic' (the default),
            'ntlm', 'cert' (ssl client certificate) or 'gssapi' (GSSAPI, Kerberos).
        cafile: A PEM file containing the CA certificates that should be trusted.

    Returns:
        The issued certificate.

    Raises:
        CouldNotRetrieveCertificateException: If something went wrong while
            fetching the cert.

    Note:
        This method is deprecated.
    """
    warnings.warn(
        DEPRECATIONWARNING,
        DeprecationWarning,
    )
    certsrv = Certsrv(server, username, password, **kwargs)
    return certsrv.get_existing_cert(req_id, encoding)


def get_ca_cert(server, username, password, encoding="b64", **kwargs):
    """
    Gets the (newest) CA certificate from a Microsoft AD Certificate Services web page.

    Args:
        server: The FQDN to a server running the Certification Authority
            Web Enrollment role (must be listening on https).
        username: The username for authentication.
        pasword: The password for authentication.
        encoding: The desired encoding for the returned certificate.
            Possible values are 'bin' for binary and 'b64' for Base64 (PEM).
        auth_method: The chosen authentication method. Either 'basic' (the default),
            'ntlm', 'cert' (ssl client certificate) or 'gssapi' (GSSAPI, Kerberos).
        cafile: A PEM file containing the CA certificates that should be trusted.

    Returns:
        The newest CA certificate from the server.

    Note:
        This method is deprecated.
    """
    warnings.warn(
        DEPRECATIONWARNING,
        DeprecationWarning,
    )
    certsrv = Certsrv(server, username, password, **kwargs)
    return certsrv.get_ca_cert(encoding)


def get_chain(server, username, password, encoding="bin", **kwargs):
    """
    Gets the chain from a Microsoft AD Certificate Services web page.

    Args:
        server: The FQDN to a server running the Certification Authority
            Web Enrollment role (must be listening on https).
        username: The username for authentication.
        pasword: The password for authentication.
        encoding: The desired encoding for the returned certificates.
            Possible values are 'bin' for binary and 'b64' for Base64 (PEM).
        auth_method: The chosen authentication method. Either 'basic' (the default),
            'ntlm', 'cert' (ssl client certificate) or 'gssapi' (GSSAPI, Kerberos).
        cafile: A PEM file containing the CA certificates that should be trusted.

    Returns:
        The CA chain from the server, in PKCS#7 format.

    Note:
        This method is deprecated.
    """
    warnings.warn(
        DEPRECATIONWARNING,
        DeprecationWarning,
    )
    certsrv = Certsrv(server, username, password, **kwargs)
    return certsrv.get_chain(encoding)


def check_credentials(server, username, password, **kwargs):
    """
    Checks the specified credentials against the specified ADCS server.

    Args:
        ca: The FQDN to a server running the Certification Authority
            Web Enrollment role (must be listening on https).
        username: The username for authentication.
        pasword: The password for authentication.
        auth_method: The chosen authentication method. Either 'basic' (the default),
            'ntlm', 'cert' (ssl client certificate) or 'gssapi' (GSSAPI, Kerberos).
        cafile: A PEM file containing the CA certificates that should be trusted.

    Returns:
        True if authentication succeeded, False if it failed.

    Note:
        This method is deprecated.
    """
    warnings.warn(
        DEPRECATIONWARNING,
        DeprecationWarning,
    )
    certsrv = Certsrv(server, username, password, **kwargs)
    return certsrv.check_credentials()



"""
DNS-01 Challenge Validator.

Implements validation logic for DNS-01 challenges according to RFC 8555.
"""

from .base import ChallengeValidator, ChallengeContext, ValidationResult


class DnsChallengeValidator(ChallengeValidator):
    """Validator for DNS-01 challenges."""

    def get_challenge_type(self) -> str:
        return "dns-01"

    def perform_validation(self, context: ChallengeContext) -> ValidationResult:
        """Perform DNS-01 challenge validation."""
        self.logger.debug("DnsChallengeValidator.perform_validation()")
        try:
            from acme2certifier.acme_srv.helper import (
                b64_url_encode,
                sha256_hash,
                txt_get,
            )
        except ImportError as e:
            return ValidationResult(
                success=False,
                invalid=True,
                error_message=f"Required dependencies not available: {e}",
                details={"import_error": str(e)},
            )

        # Handle wildcard domain
        fqdn = self._handle_wildcard_domain(context.authorization_value)

        # Construct the DNS record name
        dns_record_name = f"_acme-challenge.{fqdn}"

        # Compute expected hash
        expected_hash = b64_url_encode(
            self.logger,
            sha256_hash(self.logger, f"{context.token}.{context.jwk_thumbprint}"),
        )

        # Query DNS
        txt_records = txt_get(self.logger, dns_record_name, context.dns_servers)

        if expected_hash in txt_records:
            success = True
        else:
            success = False
            self.logger.debug(
                "DnsChallengeValidator.perform_validation(): Expected hash %s not found in DNS records: %s",
                expected_hash,
                txt_records,
            )

        self.logger.debug(
            "DnsChallengeValidator.perform_validation() ended with: %s", success
        )
        return ValidationResult(
            success=success,
            invalid=not success,
            error_message=(
                None
                if success
                else '{"status": 403, "type": "urn:ietf:params:acme:error:incorrectResponse", "detail": "DNS record not found or incorrect"}'
            ),
            details={
                "dns_record": dns_record_name,
                "expected_hash": expected_hash,
                "found_records": txt_records,
            },
        )

    def _handle_wildcard_domain(self, fqdn: str) -> str:
        """Handle wildcard domain by removing the '*.' prefix."""
        self.logger.debug(
            "DnsChallengeValidator._handle_wildcard_domain() called with: %s", fqdn
        )
        if fqdn.startswith("*."):
            fqdn = fqdn[2:]
            self.logger.debug(
                "DnsChallengeValidator._handle_wildcard_domain(): Wildcard domain detected, updated FQDN: %s",
                fqdn,
            )
        self.logger.debug(
            "DnsChallengeValidator._handle_wildcard_domain() returning: %s", fqdn
        )
        return fqdn
